@@ -1,6 +1,7 @@
 """
 Tests database tools
 """
+import csv
 import pathlib
 import time
 import unittest
@@ -25,6 +26,16 @@ def run_locked(func):
     return wrapper
 
 
+def _read_csv(file_path) -> str:
+    """Read and return the contents of a csv file"""
+    with open(file_path, newline='') as file:
+        csv_reader = csv.reader(file, delimiter=',')
+        csv_contents = ''
+        for row in csv_reader:
+            csv_contents += ','.join(row)
+    return csv_contents
+
+
 class TestDatabaseTable(TimeBasedDatabaseTable):
     @property
     def table_name(self) -> str:
@@ -39,48 +50,6 @@ class TestDatabaseTable(TimeBasedDatabaseTable):
             ('col_3', 'BOOL'),
         ]
 
-    def retrieve(self,
-                 start_time: datetime = None,
-                 end_time: datetime = None,
-                 duration: timedelta = timedelta(hours=1),
-                 ) -> List[List]:
-        if duration is not None:
-            if start_time is None and end_time is None:
-                end_time = datetime.now()
-
-            if start_time and end_time:
-                pass
-            elif start_time:
-                end_time = start_time + duration
-            elif end_time:
-                start_time = end_time - duration
-            command = self._select_query + " WHERE datetime BETWEEN :start_time AND :end_time"
-            values = {
-                'start_time': start_time,
-                'end_time': end_time,
-            }
-        else:
-            command = self._select_query
-            values = {}
-        with sqlite3.connect(self.file_path) as db:
-            with closing(db.cursor()) as cur:
-                cur.execute(command, values)
-                rows = cur.fetchall()
-        out = []
-        for row in rows:
-            # cast each value to the types they should be
-            try:
-                dt = datetime.strptime(row[0], "%Y-%m-%d %H:%M:%S.%f")
-            except ValueError:
-                dt = datetime.strptime(row[0], "%Y-%m-%d %H:%M:%S")
-            col_1 = str(row[1])
-            col_2 = int(row[2])
-            col_3 = bool(row[3])
-            # add the row of values to the output
-            r = [dt, col_1, col_2, col_3]
-            out.append(r)
-        return out
-
 
 class BadTestTimeBasedDatabaseTable(TimeBasedDatabaseTable):
     @property
@@ -95,15 +64,6 @@ class BadTestTimeBasedDatabaseTable(TimeBasedDatabaseTable):
     def table_name(self) -> str:
         return 'A_BAD_TEMPORAL_TEST_DB_TABLE'
 
-    def retrieve(self,
-                 start_time: datetime = None,
-                 end_time: datetime = None,
-                 duration: timedelta = timedelta(hours=1),
-                 ) -> List[List]:
-        # just pass this because we dont expect to be able to instantiate a class instance
-        pass
-
-
 class AnotherTestDatabaseTable(DatabaseTable):
     @property
     def column_name_types(self) -> List[Tuple[str, str]]:
@@ -119,7 +79,8 @@ class AnotherTestDatabaseTable(DatabaseTable):
 
 
 class TestDatabaseTable31DayRetentionPolicy(DatabaseRetentionPolicy):
-    """A policy to only retain entries made within the last 31 days"""
+    """A policy to only retain entries made within the last 31 days, any entries exactly at the end time will not be
+    retained"""
     @property
     def _retention_policy(self) -> str:
         policy = "delete from TEST_DB_TABLE where datetime not between :start_time and :end_time"
@@ -139,6 +100,7 @@ class TestDatabaseTable31DayRetentionPolicy(DatabaseRetentionPolicy):
 class TestDBTools(unittest.TestCase):
 
     db_path: str = 'test_db.db'
+    db_path_without_file_type: str = 'test_db'
     db_time_based_table: TimeBasedDatabaseTable = None
     db_retention_policy_31_days: TestDatabaseTable31DayRetentionPolicy = TestDatabaseTable31DayRetentionPolicy(enabled=False)
 
@@ -169,7 +131,28 @@ class TestDBTools(unittest.TestCase):
                 self.assertEqual(f'col_{col_number}', column_names[index])
 
     @run_locked
-    def test_insert_retrieve_remove(self):
+    def test_database_table_insert_retrieve(self):
+        another_db_table = AnotherTestDatabaseTable(file_path=self.db_path)
+        expected_entries = []
+        for i in range(5):
+            expected_entries.append([f'text {i}', i, True if i % 2 == 0 else False])
+            another_db_table.insert((f'text {i}', i, True if i % 2 == 0 else False))
+            retrieved = another_db_table.retrieve()
+            self.assertEqual(expected_entries, retrieved,
+                             'ensure all entries were inserted and retrieved properly')
+
+        # test inserting multiple values at once
+        expected_entries.append([f'text {5}', 5, True if 5 % 2 == 0 else False])
+        expected_entries.append([f'text {6}', 6, True if 6 % 2 == 0 else False])
+        another_db_table.insert(*[(f'text {5}', 5, True if 5 % 2 == 0 else False),
+                                  (f'text {6}', 6, True if 6 % 2 == 0 else False),
+                                  ])
+        retrieved = another_db_table.retrieve()
+        self.assertEqual(expected_entries, retrieved,
+                         'ensure inserting multiple values at once works')
+
+    @run_locked
+    def test_time_based_table_insert_retrieve_remove(self):
         wait_time = 1  # second, wait time between inserting data into the table
 
         # test inserting into the database table
@@ -270,6 +253,12 @@ class TestDBTools(unittest.TestCase):
     def test_retention_policy(self):
         self.assertFalse(self.db_retention_policy_31_days.enabled, 'retention policy should not be enabled')
 
+        self.db_retention_policy_31_days.enabled = True
+        self.assertTrue(self.db_retention_policy_31_days.enabled, 'retention policy should be enabled')
+
+        self.db_retention_policy_31_days.enabled = False
+        self.assertFalse(self.db_retention_policy_31_days.enabled, 'retention policy should not be enabled')
+
         # initially add values into the database table
         insert_1_time = datetime.now()
         insert_1 = (insert_1_time, 'insert 1', 1, True)
@@ -279,16 +268,20 @@ class TestDBTools(unittest.TestCase):
         insert_almost_31_days_ago = (insert_almost_31_days_ago_time, 'insert almost 31 days ago', -31, True)
         self.db_time_based_table.insert(insert_almost_31_days_ago)
 
+        insert_31_days_ago_time = datetime.now() - timedelta(days=31)
+        insert_31_days_ago = (insert_31_days_ago_time, 'insert 31 days ago', -31, True)
+        self.db_time_based_table.insert(insert_31_days_ago)
+
         insert_32_days_ago_time = datetime.now() - timedelta(days=32)
         insert_32_days_ago = (insert_32_days_ago_time, 'insert 32 days ago', -32, True)
         self.db_time_based_table.insert(insert_32_days_ago)
 
         # run the policy on the database table, but dont enable the policy so the values should not be removed
-        expected_1_and_almost_31_and_32 = [list(insert_1), list(insert_almost_31_days_ago), list(insert_32_days_ago)]
+        expected_1_and_almost_31_and_31_and_32 = [list(insert_1), list(insert_almost_31_days_ago), list(insert_31_days_ago), list(insert_32_days_ago)]
         self.assertFalse(self.db_retention_policy_31_days(self.db_time_based_table),
                          'policy should not run because it is not enabled')
-        retrieved_1_and_almost_31_and_32 = self.db_time_based_table.retrieve(duration=None)
-        self.assertEqual(expected_1_and_almost_31_and_32, retrieved_1_and_almost_31_and_32,
+        retrieved_1_and_almost_31_and_31_and_32 = self.db_time_based_table.retrieve(duration=None)
+        self.assertEqual(expected_1_and_almost_31_and_31_and_32, retrieved_1_and_almost_31_and_31_and_32,
                          'all entries should be in the database if duration is None')
 
         # run the policy on the database table with the policy enabled so values older than 31 days should be removed
@@ -330,4 +323,104 @@ class TestDBTools(unittest.TestCase):
                 'malicious value'
             )
 
+    @run_locked
+    def test_save_to_csv(self):
+        """test saving tables into a csv file"""
+        basic_db_table_csv_heading = 'col_1,col_2,col_3'
+        db_time_based_table_csv_heading = 'datetime,col_1,col_2,col_3'
+
+        # add a second table to the database to make sure data from different tables in the same database get saved
+        # in separate csv files
+        basic_db_table = AnotherTestDatabaseTable(file_path=self.db_path)
+        basic_db_table_csv_path = Path('test save csv.csv')
+
+        # ensure the default csv file path is set properly
+        self.assertEqual(self.db_time_based_table._csv_path,
+                         Path(f'{self.db_path_without_file_type} - {self.db_time_based_table.table_name}.csv'),
+                         'ensure the default csv file path is set properly')
+
+        # if any csv files from previous tests exist, remove them
+        if self.db_time_based_table._csv_path.exists():
+            self.db_time_based_table._csv_path.unlink()
+        if basic_db_table_csv_path.exists():
+            basic_db_table_csv_path.unlink()
+        self.assertFalse(self.db_time_based_table._csv_path.exists(),
+                         'ensure creating a csv with the default csv name works')
+        self.assertFalse(basic_db_table_csv_path.exists(),
+                         'ensure creating a csv with the default csv name works')
+
+        # ensure the initial csv files have the correct headings and file names
+        basic_db_table.save_to_csv(file_path=basic_db_table_csv_path)
+        self.db_time_based_table.save_to_csv()
+        self.assertTrue(self.db_time_based_table._csv_path.exists(),
+                        'ensure creating a csv with the default csv name works')
+        self.assertTrue(basic_db_table_csv_path.exists(),
+                        'ensure creating a csv with a custom csv name works')
+        csv_contents = _read_csv(basic_db_table_csv_path)
+        self.assertEqual(csv_contents, basic_db_table_csv_heading,
+                         'ensure the heading row for the basic db table was created correctly')
+        csv_contents = _read_csv(self.db_time_based_table._csv_path)
+        self.assertEqual(csv_contents, db_time_based_table_csv_heading,
+                         'ensure the heading row for the time based db table was created correctly')
+
+        entry_times = [datetime(year=2019, month=9, day=30, hour=i, minute=0, second=0, microsecond=0) for i in range(6)]
+
+        #  insert values into the database tables, save the tables as csv files, and ensure the contents are correct
+        expected_basic_db_table_csv_contents = basic_db_table_csv_heading
+        expected_time_based_db_table_csv_contents = db_time_based_table_csv_heading
+        for i, entry_time in enumerate(entry_times):
+            i = float(i)
+            basic_db_table.insert((f'text {i}', i, True if i % 2 == 0 else False))
+            self.db_time_based_table.insert((entry_time, f'text {i}', i, True if i % 2 == 0 else False))
+            # save the csv files
+            basic_db_table.save_to_csv(basic_db_table_csv_path)
+            self.db_time_based_table.save_to_csv()
+            # ensure the csv updates properly
+            expected_basic_db_table_csv_contents += f'text {i},{i},{1 if i % 2 == 0 else 0}'
+            expected_time_based_db_table_csv_contents += f'{datetime.strftime(entry_time, "%Y-%m-%d %H:%M:%S")},text {i},{i},{1 if i % 2 == 0 else 0}'
+            csv_contents = _read_csv(basic_db_table_csv_path)
+            self.assertEqual(csv_contents, expected_basic_db_table_csv_contents,
+                             'ensure the contents of the basic db table were saved to the csv correctly')
+            csv_contents = _read_csv(self.db_time_based_table._csv_path)
+            self.assertEqual(csv_contents, expected_time_based_db_table_csv_contents,
+                             'ensure the contents of the time based db table were saved to the csv correctly')
+
+        # test saving only certain times for the time based database table
+        self.db_time_based_table.save_to_csv(start_time=datetime(year=2019, month=9, day=30, hour=2, minute=0, second=0, microsecond=0),
+                                             end_time=datetime(year=2019, month=9, day=30, hour=4, minute=0, second=0, microsecond=0),
+                                             )
+        csv_contents = _read_csv(self.db_time_based_table._csv_path)
+        expected = "datetime,col_1,col_2,col_3" \
+                   "2019-09-30 02:00:00,text 2.0,2.0,1" \
+                   "2019-09-30 03:00:00,text 3.0,3.0,0" \
+                   "2019-09-30 04:00:00,text 4.0,4.0,1"
+        self.assertEqual(csv_contents, expected,
+                         'ensure saving a csv with a start time and end time works')
+
+        self.db_time_based_table.save_to_csv(start_time=datetime(year=2019, month=9, day=30, hour=0, minute=0, second=0, microsecond=0),
+                                             duration=timedelta(hours=3),
+                                             )
+        csv_contents = _read_csv(self.db_time_based_table._csv_path)
+        expected = "datetime,col_1,col_2,col_3" \
+                   "2019-09-30 00:00:00,text 0.0,0.0,1" \
+                   "2019-09-30 01:00:00,text 1.0,1.0,0" \
+                   "2019-09-30 02:00:00,text 2.0,2.0,1" \
+                   "2019-09-30 03:00:00,text 3.0,3.0,0"
+        self.assertEqual(csv_contents, expected,
+                         'ensure saving a csv with a start and duration works')
+
+        self.db_time_based_table.save_to_csv(end_time=datetime(year=2019, month=9, day=30, hour=4, minute=0, second=0, microsecond=0),
+                                             duration=timedelta(hours=2),
+                                             )
+        csv_contents = _read_csv(self.db_time_based_table._csv_path)
+        expected = "datetime,col_1,col_2,col_3" \
+                   "2019-09-30 02:00:00,text 2.0,2.0,1" \
+                   "2019-09-30 03:00:00,text 3.0,3.0,0" \
+                   "2019-09-30 04:00:00,text 4.0,4.0,1"
+        self.assertEqual(csv_contents, expected,
+                         'ensure saving a csv with an end time and duration works')
+
+
+
+
 
